import { CompositeLayer, Layer } from 'deck.gl';
import { LineLayer, ScatterplotLayer, TextLayer } from '@deck.gl/layers';
import { CompositeLayerType, Coordinate, LayerProps, EDGE_COLOR_DEFAULT } from '../../mapvis.types';
import { BrushingExtension, CollisionFilterExtension } from '@deck.gl/extensions';
import { scaleLinear, ScaleLinear, color, interpolateRgb } from 'd3';
import { nodeColorRGB } from '../../utils';
import { CategoricalStats, EdgeQueryResult, NodeQueryResult } from 'ts-common';

interface ColorScales {
  [label: string]: ScaleLinear<string, string>;
}

interface NodeLinkLayerState {
  colorScales: ColorScales;
  [key: string]: any;
}

export class NodeLinkLayer extends CompositeLayer<CompositeLayerType> {
  static type = 'nodelink';
  private _layers: Record<string, Layer> = {};

  state: NodeLinkLayerState = {
    colorScales: {},
  };

  constructor(props: LayerProps) {
    super(props);
  }

  updateState({ props, changeFlags }: { props: any; changeFlags: any }) {
    // TODO: Remove any here
    if (changeFlags.propsOrDataChanged) {
      const colorScales: ColorScales = {};
      const nodesSettings = props.settings?.[NodeLinkLayer.type]?.nodes;
      const nodesMetadata = props.graphMetadata?.nodes?.types;

      if (nodesSettings && nodesMetadata) {
        Object.keys(nodesSettings).forEach(label => {
          const nodeSettings = nodesSettings[label];
          const nodeDistribution = nodesMetadata[label]?.attributes;

          if (nodeSettings?.colorByAttribute && nodeDistribution) {
            const colorAttribute = nodeSettings.colorAttribute;
            const attributeData = nodeDistribution[colorAttribute];

            if (nodeSettings.colorAttributeType === 'number' && attributeData) {
              colorScales[label] = this.setNumericalColor(
                attributeData.statistics.min,
                attributeData.statistics.max,
                nodeSettings.colorScale,
              );
            }
          }
        });
      }

      const edgesSettings = props.settings?.[NodeLinkLayer.type]?.edges;
      const edgesMetadata = props.graphMetadata?.edges?.types;
      if (edgesSettings && edgesMetadata) {
        Object.keys(edgesSettings).forEach(label => {
          const edgeSettings = edgesSettings[label];
          const edgeDistribution = edgesMetadata[label]?.attributes;

          if (edgeSettings?.colorByAttribute && edgeDistribution) {
            const colorAttribute = edgeSettings.colorAttribute;
            const attributeData = edgeDistribution[colorAttribute];

            if (edgeSettings.colorAttributeType === 'number' && attributeData) {
              colorScales[label] = this.setNumericalColor(
                attributeData.statistics.min,
                attributeData.statistics.max,
                edgeSettings.colorScale,
              );
            }
          }
        });
      }

      this.setState({ colorScales });
    }

    return changeFlags.propsOrDataChanged || changeFlags.somethingChanged;
  }

  setNumericalColor(min: number, max: number, colorScale: string) {
    return scaleLinear<string>().domain([min, max]).range(['white', colorScale]).interpolate(interpolateRgb);
  }

  rgbStringToArray(rgbString: string): [number, number, number] {
    const rgb = color(rgbString);
    if (!rgb || !rgb.formatRgb()) return [0, 0, 0];
    const match = rgb.formatRgb().match(/\d+/g);
    if (!match || match.length !== 3) return [0, 0, 0];
    const [r, g, b] = match.map(Number) as [number, number, number];
    return [r, g, b];
  }

  getEdgeColor(d: EdgeQueryResult) {
    const graphMetadata = this.props.graphMetadata;
    const layerSettings = this.props.settings[NodeLinkLayer.type];

    if (layerSettings?.edges?.[d.label]?.colorByAttribute) {
      const attribute = layerSettings?.edges[d.label].colorAttribute;
      let attributeValue = d.attributes[attribute] as any;

      if (['string', 'categorical', 'boolean'].includes(layerSettings?.edges[d.label].colorAttributeType)) {
        return nodeColorRGB(
          layerSettings?.edges[d.label]?.colorMapping?.[attributeValue] ??
            (graphMetadata.edges.types[d.label]?.attributes[attribute].statistics as CategoricalStats)?.values?.indexOf(attributeValue),
        );
      } else if (layerSettings?.edges[d.label].colorAttributeType === 'number') {
        if (typeof attributeValue === 'string') {
          const numericValue = parseFloat(attributeValue.replace(/[^0-9.]/g, ''));
          if (!isNaN(numericValue)) {
            attributeValue = numericValue;
          }
        }
        const colorScale = this.state.colorScales[d.label];
        return this.rgbStringToArray(colorScale(attributeValue));
      }
    }
    return EDGE_COLOR_DEFAULT;
  }

  renderLayers() {
    const { data, settings, getNodeLocation, ml, graphMetadata, selected } = this.props;
    const layerSettings = settings[NodeLinkLayer.type];

    const brushingExtension = new BrushingExtension();
    const collisionFilter = new CollisionFilterExtension();

    const nodeLocations = data.nodes.reduce((acc: Record<string, Coordinate>, node: NodeQueryResult) => {
      const pos = getNodeLocation(node._id);
      if (pos && (pos[0] !== 0 || pos[1] !== 0)) {
        acc[node._id] = pos;
      }
      return acc;
    }, {});

    const hiddenNodes = new Set(data.nodes.filter(node => layerSettings.nodes[node.label]?.hidden).map(node => node._id));

    graphMetadata.edges.labels.forEach((label: string) => {
      const layerId = `${label}-edges-line`;

      const showArrows = layerSettings?.showArrows;

      const parseEdge = (edge: EdgeQueryResult) => {
        if (edge == null || edge?.attributes?.hidden) return [];
        if (!nodeLocations[edge.from] || !nodeLocations[edge.to]) return [];

        const [sx, sy] = nodeLocations[edge.from];
        const [tx, ty] = nodeLocations[edge.to];

        const [wsx, wsy] = this.context.viewport.projectPosition([sx, sy, 1]);
        // eslint-disable-next-line prefer-const
        let [wtx, wty, wtz] = this.context.viewport.projectPosition([tx, ty, 1]);

        if (showArrows) {
          // Perpendicular normalized vector
          let ax = wtx - wsx;
          let ay = wty - wsy;
          const amag = Math.sqrt(ax * ax + ay * ay);
          if (amag != 0) {
            ax /= amag;
            ay /= amag;
          }

          const nodeRadius = wtz * 40;

          if (wtx - wsx != 0) wtx -= (nodeRadius / amag) * (wtx - wsx);
          if (wty - wsy != 0) wty -= (nodeRadius / amag) * (wty - wsy);

          // Draw arrow heads
          const arrowSize = nodeRadius / 7;
          const arrowRatio = 1.75;

          let px = wty - wsy;
          let py = -(wtx - wsx);
          const pmag = Math.sqrt(px * px + py * py);
          if (pmag != 0) {
            px /= pmag;
            py /= pmag;
          }

          const arrow1_x = px - ax * arrowRatio;
          const arrow1_y = py - ay * arrowRatio;
          const arrow2_x = -px - ax * arrowRatio;
          const arrow2_y = -py - ay * arrowRatio;

          return [
            // Arrow head line 1
            {
              label: edge.label,
              from: this.context.viewport.unprojectPosition([wtx, wty]),
              to: this.context.viewport.unprojectPosition([wtx + arrow1_x * arrowSize, wty + arrow1_y * arrowSize]),
              attributes: edge.attributes,
            },
            // Edge
            {
              _id: edge._id,
              label: edge.label,
              attributes: edge.attributes,
              from: [sx, sy, 0] as [number, number, number],
              fromID: edge.from,
              to: this.context.viewport.unprojectPosition([wtx, wty]),
              toID: edge.to,
            },
            // Arrow head line 2
            {
              label: edge.label,
              from: this.context.viewport.unprojectPosition([wtx, wty]),
              to: this.context.viewport.unprojectPosition([wtx + arrow2_x * arrowSize, wty + arrow2_y * arrowSize]),
              attributes: edge.attributes,
            },
          ];
        } else {
          return {
            _id: edge._id,
            label: edge.label,
            attributes: edge.attributes,
            from: [sx, sy, 0] as [number, number, number],
            fromID: edge.from,
            to: [tx, ty, 0] as [number, number, number],
            toID: edge.to,
          };
        }
      };
      const edgeData = data.edges.flatMap(parseEdge).filter(edge => {
        return (
          edge != null && edge.from && edge.to && edge.fromID && edge.toID && !hiddenNodes.has(edge.fromID) && !hiddenNodes.has(edge.toID)
        );
      });

      this._layers[layerId] = new LineLayer({
        id: layerId,
        data: edgeData,
        visible: !layerSettings?.edges[label]?.hidden,
        pickable: true,
        getWidth: layerSettings?.edges[label]?.width || 1,
        getSourcePosition: d => d.from,
        getTargetPosition: d => d.to,
        getColor: (d: EdgeQueryResult) => this.getEdgeColor(d),
        extensions: [brushingExtension],
        brushingEnabled: layerSettings?.enableBrushing,
      });
    });

    if (ml?.linkPrediction?.enabled) {
      this._layers['link_prediction'] = new LineLayer({
        id: 'link-prediction-layer',
        data: ml.linkPrediction.result,
        pickable: false,
        getWidth: 1,
        getSourcePosition: d => d.from,
        getTargetPosition: d => d.to,
        getColor: d => [0, 0, 0],
      });
    }

    graphMetadata.nodes.labels.forEach((label: string) => {
      const layerId = `${label}-nodes-scatterplot`;
      const textLayerId = `${label}-label-target`;

      const nodes = data.nodes.filter(node => nodeLocations[node._id] && node.label === label);

      this._layers[layerId] = new ScatterplotLayer({
        id: layerId,
        visible: !layerSettings?.nodes[label]?.hidden,
        data: nodes,
        pickable: true,
        getFillColor: d => {
          if (layerSettings?.nodes[label].colorByAttribute) {
            const attribute = layerSettings?.nodes[label].colorAttribute;
            let attributeValue = d.attributes[attribute];
            if (['string', 'categorical', 'boolean'].includes(layerSettings?.nodes[label].colorAttributeType)) {
              return nodeColorRGB(
                layerSettings?.nodes[label]?.colorMapping[attributeValue] ??
                  (graphMetadata.nodes.types[label].attributes[attribute].statistics as CategoricalStats).values.indexOf(attributeValue),
              );
            } else if (layerSettings?.nodes[label].colorAttributeType === 'number') {
              if (typeof attributeValue === 'string') {
                const numericValue = parseFloat(attributeValue.replace(/[^0-9.]/g, ''));
                if (!isNaN(numericValue)) {
                  attributeValue = numericValue;
                }
              }
              const colorScale = this.state.colorScales[label];
              return this.rgbStringToArray(colorScale(attributeValue));
            }
          }
          return nodeColorRGB(layerSettings?.nodes[label].color);
        },
        getPosition: (d: NodeQueryResult) => getNodeLocation(d._id),
        getRadius: () => layerSettings?.nodes[label]?.size,
        radiusMinPixels: 5,
        getLineWidth: (d: NodeQueryResult) => (selected && selected.some(sel => sel._id === d._id) ? 2 : 1),
        lineWidthUnits: 'pixels',
        stroked: true,
        updateTriggers: {
          getIcon: [selected],
          getRadius: [layerSettings?.nodes[label].size],
          getFillColor: [this.state.colorScales],
        },
      });

      this._layers[textLayerId] = new TextLayer({
        id: textLayerId,
        data: nodes,
        getPosition: (d: NodeQueryResult) => getNodeLocation(d._id),
        getText: (d: NodeQueryResult) => d.label,
        getSize: (d: NodeQueryResult) => (layerSettings?.nodes[label]?.size * 2.5) / d.label.length,
        getAlignmentBaseline: 'center',
        getRadius: 10,
        radiusScale: 20,
        getColor: [255, 255, 255],
        sizeUnits: 'meters',
        sizeMaxPixels: 64,
        characterSet: 'auto',
        fontFamily: 'InterVariable',
        billboard: false,
        getAngle: () => 0,
        collisionGroup: 'textLabels',
        extensions: [collisionFilter],
        collisionEnabled: layerSettings?.collisionEnabled ?? true,
        getCollisionPriority: () => 100,
        collisionTestProps: { sizeScale: 5 },
      });
    });

    return Object.values(this._layers);
  }
}

NodeLinkLayer.layerName = 'NodeLink';
